indexing

	description: "Class that implements a simple HTTP 1.1 server."

	standards: "RFC 2616"

	known_bugs:
		"1. request URI is not decoded when of the form '%% HEX HEX'."

	not_implemented:
		"Main things not done, and probably never will:%
		%1. Chunked encoding.%
		%2. Range support.%
		%3. Persistent connections.%
		%4. Content negotation.%
		%But it's relatively easy to add them inheriting from this server%
		%I guess."

	author: "Berend de Boer"
	date: "$Date: 2004/12/18 $"
	revision: "$Revision: #8 $"


class

	EPX_HTTP_SERVER


inherit

	ANY

	EPX_CURRENT_PROCESS
		export
			{NONE} all
		end

	EPX_FILE_SYSTEM
		export
			{NONE} all;
			{ANY} is_directory
		end

	EPX_HTTP_RESPONSE
		export
			{NONE} all;
			{EPX_HTTP_SERVLET,EPX_HTTP_SERVER} is_three_digit_response
		end

	EPX_XML_ROUTINES
		export
			{EPX_HTTP_SERVLET} all
		end

	EPX_MIME_TYPE_NAMES
		export
			{NONE} all
		end

	EPX_HTTP_METHODS
		export
			{NONE} all
		end

	EPX_MIME_FIELD_NAMES
		export
			{NONE} all
		end


creation

	make,
	make_virtual


feature {NONE} -- Initialization

	do_make (a_port: INTEGER) is
			-- Initialize webserver that listens to requests on the
			-- loop-back address.
		require
			valid_port: a_port >= 0 and then a_port <= 65535
		local
			host: EPX_HOST
			service: EPX_SERVICE
			hp: EPX_HOST_PORT
		do
			create root_url.make ("file://localhost/")
			create service.make_from_port (a_port, "tcp")
			host := new_host
			create hp.make (host, service)
			create socket.listen_by_address (hp)
			socket.set_blocking_io (False)
			create resources.make
			create dynamic_resources.make
			create fixed_resources.make (128)
			create response_header.make (256)
			serve_xhtml_if_supported := True
		end

	make (a_port: INTEGER; a_root: STRING) is
			-- Initialize the web server with `a_root' as the '/' URL.
		require
			root_not_empty: a_root /= Void and then not a_root.is_empty
			root_exists: is_directory (a_root)
			valid_port: a_port >= 0 and then a_port <= 65535
		do
			create root.make_from_string (resolved_path_name (a_root))
			if
				root.item (root.count) = '/' or else
				root.item (root.count) = '/'
			then
				root.remove (root.count)
			end
			do_make (a_port)
		ensure
			open: is_open
			not_virtual: not is_virtual
		end

	make_virtual (a_port: INTEGER) is
			-- Initialize the web server with no root in the file
			-- system. It will not serve files, only servlets.
			-- Use `register_url' to add recognized URLs in that case.
		require
			valid_port: a_port >= 0 and then a_port <= 65535
		do
			root := Void
			do_make (a_port)
		ensure
			open: is_open
			virtual: is_virtual
		end

	new_host: EPX_HOST is
			-- The host name/ip address this web server listen on.
				do
			--create Result.make_from_name ("localhost")
			-- Only listen to 127.0.0.1. Web server only reachable from
			-- local machine.
			create Result.make_from_ip4_loopback
			-- Listen on all local addresses. Web server reachable by everyone.
			--create Result.make_from_ip4_any
		ensure
			host_assigned: Result /= Void
			host_found: Result.found
		end


feature -- Done

	shutdown is
			-- Don't listen to requests anymore.
		require
			open: is_open
		do
			socket.close
			socket := Void
			from
				resources.start
			until
				resources.after
			loop
				resources.item_for_iteration.shutdown
				resources.forth
			end
		ensure
			closed: not is_open
		end


feature -- Status

	has_connected_client: BOOLEAN is
			-- Is there a client whose request we currently process?
		do
			Result := client /= Void and then client.is_open
		end

	has_client_disconnected: BOOLEAN is
			-- Has client decided to disconnect?
		require
			client_connected: has_connected_client
		do
			select_readable.check_for_reading.wipe_out
			select_readable.check_for_reading.put (client)
			select_readable.set_timeout (select_timeout)
			select_readable.execute
			-- We assume the HTTP server has read everything send by the
			-- client. When it becomes readable, we assume that the
			-- connection is closed, because we can call read() to return
			-- 0 (i.e. end-of-file).
			Result := select_readable.ready_for_reading.has (client)
			select_readable.check_for_reading.wipe_out
			select_readable.ready_for_reading.wipe_out
		end

	is_open: BOOLEAN is
			-- Does the server listen on the given port?
		do
			Result := socket /= Void
		end

	is_virtual: BOOLEAN is
			-- Does this server serve entirely from servlets?
			-- If so, it will never read the local file system for files
			-- directly.
		do
			Result := root = Void
		end

	serve_xhtml_if_supported: BOOLEAN
			-- If client says it supports XHTML, send mime type XHTML.


feature -- Access

	root: STDC_PATH
			-- Directory in file system where files are served from.

	root_url: EPX_URI
			-- Used to resolve relative components

	server_name: STRING is
			-- How server is known to the client.
		once
			Result := "eposix/1.0"
		ensure
			server_name_not_empty: server_name /= Void and then not server_name.is_empty
		end


feature -- Change

	set_serve_xhtml_if_supported (a_enable: BOOLEAN) is
			-- Change `serve_xhtml_if_supported'.
		do
			serve_xhtml_if_supported := a_enable
		ensure
			serve_xhtml_if_supported_set: serve_xhtml_if_supported = a_enable
		end


feature -- Custom resources, non-files

	locate_servlet (a_path: STRING): EPX_HTTP_SERVLET is
			-- Return the servlet that handles this path or Void if there
			-- is none.
		require
			path_not_empty: a_path /= Void and then not a_path.is_empty
			path_is_absolute: a_path.item (1) = '/'
		local
			dynamic_path: EPX_HTTP_DYNAMIC_PATH
		do
			fixed_resources.search (a_path)
			if fixed_resources.found then
				Result := fixed_resources.found_item
			else
				from
					dynamic_resources.start
				until
					Result /= Void or else
					dynamic_resources.after
				loop
					dynamic_path := dynamic_resources.item_for_iteration
					if dynamic_path.matches (a_path) then
						Result := dynamic_path.resource
						from
							dynamic_path.fields.start
						until
							dynamic_path.fields.after
						loop
							-- Fields in the path are added to
							-- `request_form_fields', except if they already
							-- exist. In that case we assume the user did
							-- want to change them.
							-- All fields in the path are also made available
							-- in the form or "part <fieldname>". Because if
							-- the space in the name, they can never clash
							-- with form fields.
							if not request_form_fields.has (dynamic_path.fields.item_for_iteration.key) then
								request_form_fields.force (dynamic_path.fields.item_for_iteration, dynamic_path.fields.item_for_iteration.key)
							end
							request_form_fields.force (dynamic_path.fields.item_for_iteration, "part " + dynamic_path.fields.item_for_iteration.key)
							dynamic_path.fields.forth
						end
					end
					dynamic_resources.forth
				end
			end
		end

	register_dynamic_resource (a_path: STRING; a_servlet: EPX_HTTP_SERVLET) is
			-- Register `a_servlet' that will be called upon to provide the
			-- output when a client requests the resource `a_path'.
			-- `a_path' is a path in the form of the urlReplacement input
			-- defined in the WSDL 1.1 specification.
		require
			path_not_empty: a_path /= Void and then not a_path.is_empty
			path_is_absolute: a_path.item (1) = '/'
			servlet_not_void: a_servlet /= Void
		local
			dynamic_path: EPX_HTTP_DYNAMIC_PATH
		do
			if not resources.has (a_servlet) then
				resources.put_last (a_servlet)
			end
			create dynamic_path.make_from_url (a_path, a_servlet)
			dynamic_resources.put_last (dynamic_path)
		ensure
			servlet_registered: resources.has (a_servlet)
			dynamic_resource_added: dynamic_resources.last.resource = a_servlet
		end

	register_fixed_resource (a_path: STRING; a_servlet: EPX_HTTP_SERVLET) is
			-- Register `a_servlet' that will be called upon to provide the
			-- output when a client requests the resource `a_path'.
			-- `a_path' should be an absolute path without relative components.
		require
			path_not_empty: a_path /= Void and then not a_path.is_empty
			path_is_absolute: a_path.item (1) = '/'
			servlet_not_void: a_servlet /= Void
			new_path: not fixed_resources.has (a_path)
		do
			if not resources.has (a_servlet) then
				resources.put_last (a_servlet)
			end
			fixed_resources.put (a_servlet, a_path)
		ensure
			servlet_registered: resources.has (a_servlet)
			resource_recognized: fixed_resources.has (a_path)
			fixed_path_is_resource: fixed_resources.item (a_path) = a_servlet
		end

	dynamic_resources: DS_LINKED_LIST [EPX_HTTP_DYNAMIC_PATH]
			-- List of resources that have a fixed path.

	fixed_resources: DS_HASH_TABLE [EPX_HTTP_SERVLET, STRING]
			-- List of resources that have a fixed path.

	resources: DS_LINKED_LIST [EPX_HTTP_SERVLET]
			-- All registered resources.


feature -- Execute requests

	process_next_request is
			-- Process the next request, if any.
		require
			open: is_open
		do
			accept_next_request
			if client /= Void then
				clear_client_request_state
				parse_next_request
			end
		end


feature {NONE} -- Recognized commands

	process_delete (a_request_uri: EPX_URI) is
			-- Handle a DELETE request.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
		local
			servlet: EPX_HTTP_SERVLET
		do
			servlet := locate_servlet (a_request_uri.path)
			if servlet /= Void then
				if servlet.are_request_form_fields_valid_for_delete (request_form_fields) then
					process_delete_servlet (servlet, a_request_uri)
				else
					report_not_found (a_request_uri.path)
				end
			else
				report_forbidden (a_request_uri.path)
			end
		end

	process_delete_servlet (a_servlet: EPX_HTTP_SERVLET; a_request_uri: EPX_URI) is
			-- Handle servlet DELETE request.
		require
			servlet_not_void: a_servlet /= Void
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			deactivated_servlet: not a_servlet.is_activated
			request_form_fields_not_void: request_form_fields /= Void
			valid_data: a_servlet.are_request_form_fields_valid_for_delete (request_form_fields)
		do
			a_servlet.activate (Current, http_method_DELETE, a_request_uri.path)
			if
				send_servlet_delete_header (a_servlet) and then
				not is_body_send
			then
				a_servlet.delete_body
				send_body (a_servlet.as_uc_string)
			end
			a_servlet.deactivate
		ensure
			servlet_deactivated: not a_servlet.is_activated
		rescue
			if a_servlet.is_activated then
				a_servlet.deactivate
			end
		end

	process_get (a_request_uri: EPX_URI) is
			-- Handle a GET request.
			-- `a_request_uri' is a fully resolved path.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			request_form_fields_not_void: request_form_fields /= Void
		local
			servlet: EPX_HTTP_SERVLET
		do
			servlet := locate_servlet (a_request_uri.path)
			if servlet /= Void then
				if servlet.are_request_form_fields_valid_for_get (request_form_fields) then
					process_get_servlet (servlet, a_request_uri)
				else
					report_not_found (a_request_uri.path)
				end
			elseif not is_virtual then
				process_get_file (a_request_uri)
			else
				report_not_found (a_request_uri.path)
			end
		end

	process_get_servlet (a_servlet: EPX_HTTP_SERVLET; a_request_uri: EPX_URI) is
			-- Send GET request to `a_servlet'.
		require
			servlet_not_void: a_servlet /= Void
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			deactivated_servlet: not a_servlet.is_activated
			request_form_fields_not_void: request_form_fields /= Void
			valid_data: a_servlet.are_request_form_fields_valid_for_get (request_form_fields)
		do
			a_servlet.activate (Current, http_method_GET, a_request_uri.path)
			if
				send_servlet_get_header (a_servlet) and then
				not is_body_send
			then
				a_servlet.get_body
				send_body (a_servlet.as_uc_string)
			end
			a_servlet.deactivate
		ensure
			servlet_deactivated: not a_servlet.is_activated
		rescue
			if a_servlet.is_activated then
				a_servlet.deactivate
			end
		end

	process_get_file (a_request_uri: EPX_URI) is
			-- `an_absolute_path' is a fully resolved resource and
			-- returned to the client if it is a file in `root'.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
		local
			resource_name: STDC_PATH
			modification_time: STDC_TIME
			if_modified_since: EPX_MIME_FIELD_IF_MODIFIED_SINCE
		do
			-- Make sure `path' is relative from `root'.
			create resource_name.make_from_string (root + a_request_uri.path)

			-- Perhaps not just check, but really, really bail out when
			-- there is still a relative component left in the path?
				check
					no_relative_components: resource_name.is_resolved
				end

			if is_directory (resource_name) then
				resource_name.set_basename (root + path_separator.out + default_html_page_name)
			end

			if is_readable (resource_name) then
				if request_fields.has (field_name_if_modified_since) then
					if_modified_since ?= request_fields.item (field_name_if_modified_since)
				end
				if if_modified_since /= Void then
					create modification_time.make_from_unix_time (status (resource_name).modification_time)
					-- Many clients send DOS times
					modification_time.to_dos_seconds
					modification_time.to_utc
					-- According to RFC2616 clients must send exact
					-- strings, because of buggy servers. Well, let's add
					-- another buggy server.
					if modification_time <= if_modified_since.date_time then
						report_unmodified
						--send_file (resource_name)
					else
						send_file (resource_name)
					end
				else
					send_file (resource_name)
				end
			else
				if is_existing (resource_name) then
					report_forbidden (a_request_uri.path)
				else
					report_not_found (a_request_uri.path)
				end
			end
		end

	process_head (a_request_uri: EPX_URI) is
			-- Handle a HEAD request.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
		local
			servlet: EPX_HTTP_SERVLET
		do
			servlet := locate_servlet (a_request_uri.path)
			if servlet /= Void then
				if servlet.are_request_form_fields_valid_for_head (request_form_fields) then
					process_head_servlet (servlet, a_request_uri)
				else
					report_not_found (a_request_uri.path)
				end
			elseif not is_virtual then
				process_head_file (a_request_uri)
			else
				report_not_found (a_request_uri.path)
			end
		end

	process_head_servlet (a_servlet: EPX_HTTP_SERVLET; a_request_uri: EPX_URI) is
			-- Handle servlet HEAD request.
		require
			servlet_not_void: a_servlet /= Void
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			deactivated_servlet: not a_servlet.is_activated
			request_form_fields_not_void: request_form_fields /= Void
			valid_data: a_servlet.are_request_form_fields_valid_for_head (request_form_fields)
		do
			a_servlet.activate (Current, http_method_HEAD, a_request_uri.path)
			send_servlet_head_header (a_servlet)
			a_servlet.deactivate
		ensure
			servlet_deactivated: not a_servlet.is_activated
		rescue
			if a_servlet.is_activated then
				a_servlet.deactivate
			end
		end

	process_head_file (a_request_uri: EPX_URI) is
			-- `an_absolute_path' is a fully resolved resource. Status is
			-- returned to the client if it is a file in `root'.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
		local
			resource_name: STDC_PATH
			last_modified: STDC_TIME
			stat: ABSTRACT_STATUS_PATH
		do
			-- Make sure `path' is relative from `root'.
			create resource_name.make_from_string (root + a_request_uri.path)

			-- Perhaps not just check, but really, really bail out when
			-- there is still a relative component left in the path?
				check
					no_relative_components: resource_name.is_resolved
				end

			if is_readable (resource_name) then
				-- Parse MIME header here
				-- Handle:
				-- Accept-Encoding if present
				write_response_code (200)
				stat := status (resource_name)
				add_field (field_name_content_length, stat.size.out)
				create last_modified.make_from_unix_time (stat.modification_time)
				last_modified.to_utc
				add_field (field_name_last_modified, last_modified.rfc_date_string)
				write_mime_header (file_name_to_mime_type (resource_name))
				-- And blank line to separate body
				add_new_line
				send_header
			else
				if is_existing (resource_name) then
					report_forbidden (a_request_uri.path)
				else
					report_not_found (a_request_uri.path)
				end
			end
		end

	process_post (a_request_uri: EPX_URI) is
			-- Handle a POST request.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			have_request_body: request_body /= Void
			request_form_fields_not_void: request_form_fields /= Void
		local
			servlet: EPX_HTTP_SERVLET
		do
			servlet := locate_servlet (a_request_uri.path)
			if servlet /= Void then
				if servlet.are_request_form_fields_valid_for_post (request_form_fields) then
					process_post_servlet (servlet, a_request_uri)
				else
					report_not_found (a_request_uri.path)
				end
			else
				report_forbidden (a_request_uri.path)
			end
		end

	process_post_servlet (a_servlet: EPX_HTTP_SERVLET; a_request_uri: EPX_URI) is
			-- Handle servlet POST request.
		require
			servlet_not_void: a_servlet /= Void
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			deactivated_servlet: not a_servlet.is_activated
			request_form_fields_not_void: request_form_fields /= Void
			have_request_body: request_body /= Void
			valid_data: a_servlet.are_request_form_fields_valid_for_post (request_form_fields)
		do
			a_servlet.activate (Current, http_method_POST, a_request_uri.path)
			if
				send_servlet_post_header (a_servlet) and then
				not is_body_send
			then
				a_servlet.post_body
				send_body (a_servlet.as_uc_string)
			end
			a_servlet.deactivate
		ensure
			servlet_deactivated: not a_servlet.is_activated
		rescue
			if a_servlet.is_activated then
				a_servlet.deactivate
			end
		end

	process_put (a_request_uri: EPX_URI) is
			-- Handle a PUT request.
		require
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			have_request_body: request_body /= Void
			request_form_fields_not_void: request_form_fields /= Void
		local
			servlet: EPX_HTTP_SERVLET
		do
			servlet := locate_servlet (a_request_uri.path)
			if servlet /= Void then
				if servlet.are_request_form_fields_valid_for_put (request_form_fields) then
					process_put_servlet (servlet, a_request_uri)
				else
					report_not_found (a_request_uri.path)
				end
			else
				report_forbidden (a_request_uri.path)
			end
		end

	process_put_servlet (a_servlet: EPX_HTTP_SERVLET; a_request_uri: EPX_URI) is
			-- Handle servlet POST request.
		require
			servlet_not_void: a_servlet /= Void
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			deactivated_servlet: not a_servlet.is_activated
			request_form_fields_not_void: request_form_fields /= Void
			have_request_body: request_body /= Void
			valid_data: a_servlet.are_request_form_fields_valid_for_put (request_form_fields)
		do
			a_servlet.activate (Current, http_method_POST, a_request_uri.path)
			if
				send_servlet_put_header (a_servlet) and then
				not is_body_send
			then
				a_servlet.put_body
				send_body (a_servlet.as_uc_string)
			end
			a_servlet.deactivate
		ensure
			servlet_deactivated: not a_servlet.is_activated
		rescue
			if a_servlet.is_activated then
				a_servlet.deactivate
			end
		end


feature {NONE} -- GET helper

	file_name_to_mime_type (a_file_name: STRING): STRING is
			-- Try to get some suitable mime type. Needs to be
			-- rewritten. Doesn't look like the way to go.
			-- Perhaps parse/use Apache's /etc/mime.types?
			-- If not recognized, `a_file_name' will be served as
			-- text/hmtl or application/xhtml+xml.
		require
			a_file_name_not_empty: a_file_name /= Void and then not a_file_name.is_empty
		local
			file_name: STDC_PATH
		do
			create file_name.make_from_string (a_file_name)
			file_name.parse (<<once_css, once_png, once_xhtml, once_xml, once_xsl, once_xslt, once_js>>)
			if file_name.suffix.is_equal (once_css) then
				Result := mime_type_text_css
			elseif file_name.suffix.is_equal (once_js) then
				Result := mime_type_text_javascript
			elseif file_name.suffix.is_equal (once_png) then
				Result := mime_type_image_png
			elseif file_name.suffix.is_equal (once_xhtml) then
				Result := mime_type_application_xhtml_plus_xml
			elseif
				file_name.suffix.is_equal (once_xml) or else
				file_name.suffix.is_equal (once_xsl) or else
				file_name.suffix.is_equal (once_xslt)
			then
				Result := mime_type_application_xml
			end
		end

	send_file (a_file_name: STRING) is
			-- Send the file `a_file_name' to `client'.
			-- Errors are ignored.
		require
			client_connected: has_connected_client
			file_name_not_empty: a_file_name /= Void and then not a_file_name.is_empty
		local
			retried: BOOLEAN
			file: EPX_FILE_DESCRIPTOR
			last_modified: STDC_TIME
		do
			if not retried then
				-- Parse MIME header here
				-- Handle:
				-- Accept-Encoding if present
				create file.open_read (a_file_name)
				write_response_code (200)
				add_field (field_name_content_length, file.status.size.out)
				create last_modified.make_from_unix_time (file.status.modification_time)
				last_modified.to_utc
				add_field (field_name_last_modified, last_modified.rfc_date_string)
				write_mime_header (file_name_to_mime_type (a_file_name))
				-- And blank line to separate body
				add_new_line
				send_header
				if client.is_open then
					client.append (file)
				end
				file.close
			end
		rescue
			if exceptions.is_developer_exception then
				if not retried then
					retried := True
					retry
				end
			end
		end

	send_servlet_get_header (servlet: EPX_HTTP_SERVLET): BOOLEAN is
			-- Was asking servlet to prepare header and sending it to the
			-- client successful?
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			servlet_not_void: servlet /= Void
		local
			retried: INTEGER
			error_message: STRING
		do
			inspect retried
			when 0 then
				servlet.get_header
				if not is_header_send then
					if not response_header_ends_with_body_separator then
						add_new_line
					end
					if has_client_disconnected then
						client.close
					else
						send_header
					end
				end
				Result := client.is_open
			when 1 then
				if not is_header_send then
					report_internal_server_error (error_message)
				end
			else
				Result := False
			end
		ensure
			header_send: Result implies is_header_send
			still_open: Result implies client.is_open
		rescue
			if exceptions.is_developer_exception then
				retried := retried + 1
				error_message := exceptions.developer_exception_name
				if is_header_send then
					is_header_send := False
					retried := 2
				end
				retry
			end
		end

	send_servlet_head_header (servlet: EPX_HTTP_SERVLET) is
			-- Send header to client, ignore exceptions that result from
			-- writing to a client that has disconnected.
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			servlet_not_void: servlet /= Void
		local
			retried: INTEGER
			error_message: STRING
		do
			inspect retried
			when 0 then
				servlet.head_header
				if not is_header_send then
					if not response_header_ends_with_body_separator then
						add_new_line
					end
					if has_client_disconnected then
						client.close
					else
						send_header
					end
				end
			when 1 then
				if not is_header_send then
					report_internal_server_error (error_message)
				end
			else
				-- exit
			end
		rescue
			if exceptions.is_developer_exception then
				retried := retried + 1
				error_message := exceptions.developer_exception_name
				if is_header_send then
					is_header_send := False
					retried := 2
				end
				retry
			end
		end


feature {NONE} -- DELETE/POST/PUT helper

	send_servlet_delete_header (servlet: EPX_HTTP_SERVLET): BOOLEAN is
			-- Was asking servlet to prepare header and sending it to the
			-- client successful?
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			servlet_not_void: servlet /= Void
		local
			retried: INTEGER
			error_message: STRING
		do
			inspect retried
			when 0 then
				servlet.delete_header
				if not is_header_send then
					if not response_header_ends_with_body_separator then
						add_new_line
					end
					send_header
				end
				Result := client.is_open
			when 1 then
				if not is_header_send then
					report_internal_server_error (error_message)
				end
			else
				Result := False
			end
		ensure
			header_send: Result implies is_header_send
			still_open: Result implies client.is_open
		rescue
			if exceptions.is_developer_exception then
				retried := retried + 1
				error_message := exceptions.developer_exception_name
				if is_header_send then
					is_header_send := False
					retried := 2
				end
				retry
			end
		end

	send_servlet_post_header (servlet: EPX_HTTP_SERVLET): BOOLEAN is
			-- Was asking servlet to prepare header and sending it to the
			-- client successful?
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			servlet_not_void: servlet /= Void
		local
			retried: INTEGER
			error_message: STRING
		do
			inspect retried
			when 0 then
				servlet.post_header
				if not is_header_send then
					if not response_header_ends_with_body_separator then
						add_new_line
					end
					send_header
				end
				Result := client.is_open
			when 1 then
				if not is_header_send then
					report_internal_server_error (error_message)
				end
			else
				Result := False
			end
		ensure
			header_send: Result implies is_header_send
			still_open: Result implies client.is_open
		rescue
			if exceptions.is_developer_exception then
				retried := retried + 1
				error_message := exceptions.developer_exception_name
				if is_header_send then
					is_header_send := False
					retried := 2
				end
				retry
			end
		end

	send_servlet_put_header (servlet: EPX_HTTP_SERVLET): BOOLEAN is
			-- Was asking servlet to prepare header and sending it to the
			-- client successful?
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			servlet_not_void: servlet /= Void
		local
			retried: INTEGER
			error_message: STRING
		do
			inspect retried
			when 0 then
				servlet.put_header
				if not is_header_send then
					if not response_header_ends_with_body_separator then
						add_new_line
					end
					send_header
				end
				Result := client.is_open
			when 1 then
				if not is_header_send then
					report_internal_server_error (error_message)
				end
			else
				Result := False
			end
		ensure
			header_send: Result implies is_header_send
			still_open: Result implies client.is_open
		rescue
			if exceptions.is_developer_exception then
				retried := retried + 1
				error_message := exceptions.developer_exception_name
				if is_header_send then
					is_header_send := False
					retried := 2
				end
				retry
			end
		end


feature {NONE} -- Execute requests

	accept_next_request is
			-- Accept the next request, if there is one.
		require
			open: is_open
		do
			client := socket.accept
			if client /= Void then
				-- At least on Windows, the socket is
				-- non-blocking. Perhaps it inherits this property from
				-- the parent.
				client.set_blocking_io (True)
				client.set_continue_on_error
				debug ("httpd")
					stderr.put_string (client.remote_address.address.out)
					stderr.write_character (':')
					stderr.put_string (client.remote_address.port.out)
					stderr.put_string (" => ")
					stderr.put_string (client.local_address.address.out)
					stderr.write_character (':')
					stderr.put_string (client.local_address.port.out)
					stderr.write_character ('%N')
				end
			end
		ensure
			client_does_not_raise_exceptions: client /= Void implies not client.raise_exception_on_error
		end

	parse_next_request is
			-- Parse the request in `client'.
			-- Previous client state should have been cleared by a call
			-- to `clear_client_request_state'.
		require
			client_open: client /= Void and then client.is_open
			response_header_empty: response_header.is_empty
			header_not_send: not is_header_send
			body_not_send: not is_body_send
			request_fields_unset: request_fields = Void
			request_form_fields_unset: request_form_fields = Void
		local
			request: STRING
			method: STRING
			ar: ARRAY [STRING]
			request_uri: EPX_URI
			mime_parser: EPX_MIME_PARSER
			http_09_request: BOOLEAN
			has_body: BOOLEAN
		do
			request := read_request_line
			ar := sh.split_on (request, ' ')
			-- I love error checking...
			if ar.count >= 2 and then ar.count <= 3 then
				method := ar.item (ar.lower)
				request_uri := make_request_uri (ar.item (ar.lower + 1))
				if ar.count = 2 then
					client_http_version := "0.9"
					http_09_request := True
				else
					client_http_version := ar.item (ar.upper)
				end
				if request_uri /= Void then
					--dump_request
					if http_09_request then
						create request_fields.make_default
						query_to_form_fields (request_uri.query)
						process_method (method, request_uri)
					else
						create mime_parser.make_from_stream (client)
						mime_parser.parse_header
						if not mime_parser.syntax_error then
							request_fields := mime_parser.part.header.fields
							debug ("httpd")
								stderr.write_string (field_name_user_agent)
								stderr.write_string (": ")
								if user_agent /= Void then
									stderr.write_string (user_agent.value)
								else
									stderr.write_string ("<not set>")
								end
								stderr.write_string ("%N%N")
							end
							-- Host header field is required for HTTP/1.1.
							if mime_parser.part.header.has (field_name_host) then
								-- Check if we know the Transfer-Encoding
								if
									not mime_parser.part.header.has (field_name_transfer_encoding) or else
									mime_parser.part.header.item (field_name_transfer_encoding).value.is_equal (once_identity)
								then
									has_body := method.is_equal (http_method_POST) or else method.is_equal (http_method_PUT)
									if
										not has_body or else
										(mime_parser.part.header.has (field_name_content_length) and then
										mime_parser.part.header.item (field_name_content_length).value.is_integer)
									then
										if has_body then
											-- I have to check here for a Expect
											-- request-header with the 100-continue
											-- expectation. If so, I must respond
											-- with the 100 continue status.
											if mime_parser.part.header.has (field_name_expect) then
												if mime_parser.part.header.item (field_name_expect).value.as_lower.is_equal (once_100_continue) then
													send_expect_continue
												else
													report_expectation_failed (mime_parser.part.header.item (field_name_expect).value)
												end
											end
											mime_parser.parse_body (mime_parser.part.header.item (field_name_content_length).value.to_integer)
											request_body := mime_parser.part.body
											if mime_parser.part.header.content_type /= Void then
												if mime_parser.part.header.content_type.value.is_equal (mime_type_application_x_www_form_urlencoded) then
													urlencoded_to_form_fields
												elseif mime_parser.part.header.content_type.value.is_equal (mime_type_multipart_form_data) then
													mime_body_to_form_fields
												end
											end
										else
											query_to_form_fields (request_uri.query)
										end
										-- Browsers can't send PUT/DELETE requests.
										-- We change `method' here if the
										-- variable http-method:XXXX exists.
										method := remap_http_method (method)
										process_method (method, request_uri)
									else
										report_content_length_required
									end
								else
									report_not_implemented (mime_parser.part.header.item (field_name_transfer_encoding).value)
								end
							else
								report_bad_request ("Field '" + field_name_host + "' not present in header (required for HTTP 1.1 requests)")
							end
						else
							report_bad_request (request)
						end
					end
				else
					report_bad_request (request)
				end
			else
				report_bad_request (request)
			end
			-- According to RFC 2616 8.2.3, if there is no 100-continue
			-- expection, the connection must not be closed until the
			-- entire request body has been read or until the client
			-- closes the connection. Otherwise, the client
			-- might not reliably receive the response message.
			-- I think my code reads the entire body, if the header and
			-- body are well-formed.
			if client.is_open then
				-- Windows doesn't a shutdown when closing the descriptor,
				-- so help Bill.
				client.shutdown_read_write
				client.close
			end
			client := Void
		end

	read_request_line: STRING is
			-- Read Request-Line from client.
		require
			client_open: client /= Void and then client.is_open
		do
			-- Follow RFC 2616 4.1 in skipping initial empty lines.
			from
				client.read_line
				create Result.make_from_string (client.last_string)
			until
				not client.errno.is_ok or else
				client.eof or else
				not Result.is_empty
			loop
				client.read_line
				create Result.make_from_string (client.last_string)
			end
			debug ("httpd")
				stderr.write_string ("====================%N")
				stderr.write_string (client.last_string)
				stderr.write_character ('%N')
			end
		ensure
			request_line_not_void: Result /= Void
		end

	process_method (a_method: STRING; a_request_uri: EPX_URI) is
			-- Call proper routine when `a_method' is an implemented
			-- method.
		require
			method_not_empty: a_method /= Void and then not a_method.is_empty
			request_uri_valid: a_request_uri /= Void and then a_request_uri.is_path_resolved and then a_request_uri.has_absolute_path
			request_fields_set: request_fields /= Void
		do
			-- As we have parsed all input, this is a good point to say
			-- that subsequent output to the client can be non-blocking.
			-- Yeah, but why??? It only complicates our end. We surely
			-- don't want to retry if the client doesn't read fast
			-- enough.
			--client.set_blocking_io (False)

			-- dispatch
			if a_method.is_equal (http_method_GET) then
				process_get (a_request_uri)
			elseif a_method.is_equal (http_method_POST) then
				if request_body /= Void then
					process_post (a_request_uri)
				else
					report_bad_request ("request body expected")
				end
			elseif a_method.is_equal (http_method_PUT) then
				if request_body /= Void then
					process_put (a_request_uri)
				else
					report_bad_request ("request body expected")
				end
			elseif a_method.is_equal (http_method_HEAD) then
				process_head (a_request_uri)
			elseif a_method.is_equal (http_method_DELETE) then
				process_delete (a_request_uri)
			else
				report_not_implemented (a_method)
			end
		end


feature {EPX_HTTP_SERVLET} -- Execute request implementation

	make_request_uri (raw_request_uri: STRING): EPX_URI is
			-- `raw_request_uri' is the URI send by the client. If it does
			-- not conform to the specs, Void is returned, else an
			-- absolute URI is returned.
		require
			request_uri_not_empty: raw_request_uri /= Void and then not raw_request_uri.is_empty
		local
			retried: BOOLEAN
			s: STRING
		do
			if
				not retried and then
				not url_encoder.has_excluded_characters (raw_request_uri)
			then
				create Result.make (raw_request_uri)
				if Result.path = Void then
					Result.set_path ("/")
				end
				if Result.has_absolute_path then
					-- Make `s' relative from `root_url'
					create s.make_from_string (raw_request_uri)
					if s.item (1) = '/' then
						s.remove (1)
					end
					create Result.make_resolve (root_url, s)
					Result.unescape_components
				else
					-- error case: path given is not absolute
					Result := Void
				end
			else
				Result := Void
			end
		ensure
			absolute_path:
				Result = Void or else
				(Result.is_path_resolved and then Result.has_absolute_path)
		rescue
			if not retried and then exceptions.is_developer_exception then
				-- We can get an exception for a really dirty URL.
				retried := True
				retry
			end
		end


feature -- Connected client state

	clear_client_request_state is
			-- Reset state variables build/maintained when handling a
			-- request in `parse_next_request'.
		do
			client_http_version := Void
			request_fields := Void
			response_header.wipe_out
			request_form_fields := Void
			request_body := Void
			is_header_send := False
			is_body_send := False
			response_code := 0
		ensure
			request_fields_unset: request_fields = Void
			request_form_fields_unset: request_form_fields = Void
			response_header_empty: response_header.is_empty
			header_not_send: not is_header_send
			body_not_send: not is_body_send
			no_response_code: response_code = 0
		end

	client: ABSTRACT_TCP_SOCKET
			-- Socket to last connected client, or void if no client connected.

	client_http_version: STRING
			--- Client HTTP version.

	is_body_send: BOOLEAN
			-- Is a complete body send to `client'?

	is_header_send: BOOLEAN
			-- Is `response_header' send to `client'?

	request_form_fields: DS_HASH_TABLE [EPX_KEY_VALUE, STRING]
			-- If the request body appears to be form data, the data is
			-- extracted into this table.
			-- For dynamic paths the parts are added to this structure
			-- prefixed with "part ". If the part is not send as a form
			-- field, it is also added as a normal form field. So parts
			-- could be listed twice.

	request_body: EPX_MIME_BODY
			-- Request body if defined for the method.

	request_fields: DS_HASH_TABLE [EPX_MIME_FIELD, STRING]
			-- Any request header fields send by the client.

	response_code: INTEGER
			-- Response code sent to client.

	response_header: STRING
			-- Header that is filled by various routines.
			-- Use `send_header' to output it to the client.

	response_header_ends_with_body_separator: BOOLEAN is
			-- Does `response_header' end with the header/body separator?
		do
			Result :=
				response_header.count >= 2 and then
				response_header.item (response_header.count) = '%N' and then
				response_header.item (response_header.count - 1) = '%R' and then
				((response_header.count = 2) or else
				 (response_header.count >= 4) and then
				 (response_header.item (response_header.count - 2) = '%N') and then
				 (response_header.item (response_header.count - 3) = '%R'))
		end

	user_agent: EPX_MIME_FIELD is
			-- Not void, if User-Agent field present.
		require
			have_request_fields: request_fields /= Void
		do
			request_fields.search (field_name_user_agent)
			if request_fields.found then
				Result := request_fields.found_item
			end
		ensure
			definition: request_fields.has (field_name_user_agent) = (Result /= Void)
		end


feature {EPX_HTTP_SERVLET} -- Redirects

	respond_see_other (a_location: STRING) is
			-- Send a 303 (or 302 for older clients) after a successful
			-- PUT/POST/DELETE.
		require
			a_location_not_empty: a_location /= Void and then not a_location.is_empty
		do
			-- For 1.0 or earlier clients a 302 seems to work better.
			if client_http_version >= "1.1" then
				write_response_code (303)
			else
				write_response_code (302)
			end
			write_mime_header (Void)
			add_field (field_name_location, a_location)
			add_field (field_name_content_length, "0")
		end


feature {EPX_HTTP_SERVLET} -- Report errors to client

	error_reasons: DS_HASH_TABLE [STRING, INTEGER] is
			-- Slightly more verbose errors than the reason phrase.
		once
			create Result.make (16)
			Result.put ("Request not understood due to malformed syntax", 400)
			Result.put ("Access to this resource is not allowed", 403)
			Result.put ("Resource not found", 404)
			Result.put ("HTTP method not allowed", 405)
			Result.put ("Content-Length field required", 411)
			Result.put ("Internal server error", 500)
			Result.put ("HTTP method or Transfer-Encoding not implemented", 501)
			Result.put ("Timeout waiting for response to request", 504)
		ensure
			error_reasons_not_void: Result /= Void
		end

	report_bad_request (a_request: STRING) is
			-- Return 400 to client.
			-- The request could not be understood by the server due to
			-- malformed syntax. The client SHOULD NOT repeat the request
			-- without modifications.
			-- The bad request, if any , is in `client'.`last_string'.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		do
			write_report (400, a_request)
		ensure
			body_send: is_body_send
		end

	report_conflict (a_message: STRING) is
			-- Return 409 to client.
			-- The request could not be completed due to a conflict with
			-- the current state of the resource. This code is only
			-- allowed in situations where it is expected that the user
			-- might be able to resolve the conflict and resubmit the
			-- request. The response body SHOULD include enough
			-- information for the user to recognize the source of the
			-- conflict.  Ideally, the response entity would include
			-- enough information for the user or user agent to fix the
			-- problem; however, that might not be possible and is not
			-- required.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		do
			write_report (409, a_message)
		ensure
			body_send: is_body_send
		end

	report_content_length_required is
			-- The server refuses to accept the request without a defined
			-- Content-Length. The client MAY repeat the request if it
			-- adds a valid Content-Length header field containing the
			-- length of the message-body in the request message.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		do
			write_report (411, Void)
		ensure
			body_send: is_body_send
		end

	report_expectation_failed (an_expectation: STRING) is
			-- The expectation given in an Expect request-header field
			-- could not be met by this server.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		do
			write_report (417, an_expectation)
		ensure
			body_send: is_body_send
		end

	report_forbidden (a_file: STRING) is
			-- Return 403 to the client.
			-- The server understood the request, but is refusing to
			-- fulfill it.  Authorization will not help and the request
			-- SHOULD NOT be repeated.
		require
			client_connected: has_connected_client
			file_not_empty: a_file /= Void and then not a_file.is_empty
			response_header_empty: response_header.is_empty
		do
			write_report (403, a_file)
		ensure
			body_send: is_body_send
		end

	report_gateway_timeout is
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
			header_not_send: not is_header_send
		do
			response_header.wipe_out
			write_report (504, Void)
		ensure
			body_send: is_body_send
		end

	report_internal_server_error (a_message: STRING) is
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
			header_not_send: not is_header_send
		do
			response_header.wipe_out
			write_report (500, a_message)
		ensure
			body_send: is_body_send
		end

	report_method_not_allowed (a_method, a_allowed_methods: STRING) is
			-- The method specified in the Request-Line is not allowed
			-- for the resource identified by the Request-URI. The
			-- response MUST include an Allow header containing a list of
			-- valid methods for the requested resource.
		require
			client_connected: has_connected_client
			a_method_not_empty: a_method /= Void and then not a_method.is_empty
			a_allowed_methods_not_void: a_allowed_methods /= Void
			response_header_empty: response_header.is_empty
		do
			write_report_header (405)
			add_field (field_name_allow, a_allowed_methods)
			write_report_body (405, a_method)
		ensure
			body_send: is_body_send
		end

	report_not_found (a_uri: STRING) is
			-- Return 404 to the client.
			-- The server has not found anything matching the Request-URI.
		require
			client_connected: has_connected_client
			file_not_empty: a_uri /= Void and then not a_uri.is_empty
			response_header_empty: response_header.is_empty
		do
			write_report (404, a_uri)
		ensure
			body_send: is_body_send
		end

	report_not_implemented (a_method: STRING) is
			-- Return 501 to client.
			-- The server does not support the functionality required to
			-- fulfill the request. This is the appropriate response when
			-- the server does not recognize the request method and is
			-- not capable of supporting it for any resource.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		do
			write_report (501, a_method)
		ensure
			body_send: is_body_send
		end

	report_unmodified is
			-- Resource is not modified, respond with 304.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
		local
			time: STDC_TIME
		do
			write_response_code (304)
			create time.make_from_now
			time.to_utc
			add_field (field_name_date, time.rfc_date_string)
			add_field (field_name_connection, once_close)
			-- And blank line to separate body
			add_new_line
			send_header
		end


	write_report (a_code: INTEGER; a_parameter: STRING) is
			-- Send an error report to `client'.
		require
			client_connected: has_connected_client
			header_not_send: not is_header_send
			response_header_empty: response_header.is_empty
			three_digit_response: is_three_digit_response (a_code)
		do
			write_report_header (a_code)
			write_report_body (a_code, a_parameter)
		ensure
			body_send: is_body_send
		end

	write_report_header (a_code: INTEGER) is
			-- Generate header of an error report to `client'.
			-- Header will be send by `write_report_body'.
		require
			client_connected: has_connected_client
			response_header_empty: response_header.is_empty
			three_digit_response: is_three_digit_response (a_code)
		do
			write_response_code (a_code)
			write_mime_header (Void)
		end

	write_report_body (a_code: INTEGER; a_parameter: STRING) is
			-- Assumes header is in `response_header' and not yet send.
		require
			client_connected: has_connected_client
			three_digit_response: is_three_digit_response (a_code)
			header_not_send: not is_header_send
		local
			reason: STRING
			xhtml: EPX_XHTML_WRITER
		do
			if not response_header_ends_with_body_separator then
				add_new_line
			end
			send_header
			-- No errors when returning header?
			if client.is_open then
				if error_reasons.has (a_code) then
					reason := error_reasons.item (a_code)
				else
					reason := reason_phrase (a_code)
					if reason.is_empty then
						reason := a_code.out
					end
				end
				create xhtml.make
				xhtml.doctype_strict
				xhtml.b_html
				xhtml.b_head
				xhtml.title (reason)
				xhtml.e_head
				xhtml.b_body
				xhtml.b_p
				xhtml.add_data (reason)
				if a_parameter /= Void then
					if has_invalid_control_characters (a_parameter) then
						xhtml.add_data (": [passed parameter has invalid control characters]")
					else
						xhtml.add_data (": ")
						if a_parameter.has ('%N') then
							xhtml.e_p
							xhtml.b_pre
							xhtml.add_data (a_parameter)
							xhtml.e_pre
							xhtml.b_p
						else
							xhtml.add_data (a_parameter)
						end
					end
				end
				if
					(a_parameter = Void) or else
					(a_parameter.count > 0 and then
					 a_parameter.item (a_parameter.count) /= '.')
				then
					xhtml.add_data (".")
				end
				xhtml.e_p
				xhtml.e_body
				reach_ie_error_treshold (xhtml)
				xhtml.e_html
				send_body (xhtml.as_uc_string)
			else
				-- Fake this for postcondition
				is_body_send := True
			end
		ensure
			body_send: is_body_send
		end


feature {EPX_HTTP_SERVLET,EPX_HTTP_SERVER} -- Response and headers to client

	write_mime_header (content_type: STRING) is
			-- Add default mime header fields for this server. Result is
			-- not written immediately to client. Call `send_header' to do that.
		require
			client_connected: has_connected_client
			header_not_ended: not response_header_ends_with_body_separator
		local
			time: STDC_TIME
		do
			add_field (field_name_server, server_name)
			create time.make_from_now
			time.to_utc
			add_field (field_name_date, time.rfc_date_string)
			add_field (field_name_mime_version, once_mime_version)
			-- text/html works for all browsers
			-- but I rather want to serve application/xhtml+xml
			-- Test if client accepts that and if it isn't Opera.
			-- Opera 7 says it accepts it, but doesn't execute JavaScript.
			-- It's up to the client to actually supply XHTML in the
			-- served files and servlets, but it's a great debugging help.
			if content_type /= Void and then not content_type.is_empty then
				add_field (field_name_content_type, content_type)
			else
				if request_fields /= Void and then request_fields.has (field_name_accept) and then request_fields.item (field_name_accept).value.has_substring (mime_type_application_xhtml_plus_xml) then
					if
						serve_xhtml_if_supported and then
						user_agent /= Void and then
						user_agent.value.substring_index (once_opera_7, 1) = 0
					then
						add_field (field_name_content_type, mime_type_application_xhtml_plus_xml)
					else
						add_field (field_name_content_type, mime_type_text_html)
					end
				else
					add_field (field_name_content_type, mime_type_text_html)
				end
			end
			add_field (field_name_accept_ranges, once_none)
			add_field (field_name_connection, once_close)
		end

	write_response_code (a_code: INTEGER) is
			-- Add HTTP response status line to `response_header'.
		require
			response_header_empty: response_header.is_empty
			three_digit_response: is_three_digit_response (a_code)
		local
			s: STRING
		do
			response_code := a_code
			response_header.append_string (once_http_version)
			response_header.append_string (a_code.out)
			s := reason_phrase (a_code)
			if not s.is_empty then
				response_header.append_character (' ')
				response_header.append_string (s)
			end
			add_new_line
		ensure
			response_code_added: not response_header.is_empty
			response_code_set: response_code = a_code
		end


feature {EPX_HTTP_SERVLET} -- Client output helpers

	ie_error_treshold: INTEGER is 512

	reach_ie_error_treshold (xml: EPX_XML_WRITER) is
			-- Horror, horror, overcome IE ErrorTresholds...
			-- You need this if the body of a response is less than
			-- 128..512 bytes and you have a response code >= 400 and you
			-- want to support IE.
		require
			header_send: is_header_send
		local
			a_body: STRING
			s: STRING
		do
			if response_code >= 400 then
				a_body := xml.unfinished_xml
				if a_body /= Void and then not a_body.is_empty then
					if a_body.count < ie_error_treshold then
						create s.make_filled ('X', ie_error_treshold - a_body.count)
						xml.add_comment (s)
					end
				end
			end
		end

	send_body (a_body: STRING) is
			-- Assume `a_body' is the entire body.
			-- If there is an error sending `s' to the client, the
			-- connection will be closed.
		require
			client_connected: has_connected_client
			header_send: is_header_send
		do
			if has_client_disconnected then
				client.close
			else
				send_string (a_body)
			end
			is_body_send := True
		ensure
			body_send: is_body_send
		end

	send_expect_continue is
			-- Immediately send a 100 Continue response to the client.
			-- If there is an error sending `s' to the client, the
			-- connection will be closed.
		require
			response_header_empty: response_header.is_empty
			client_connected: has_connected_client
		do
			write_response_code (100)
			send_string (response_header)
			response_header.wipe_out
		ensure
			response_header_empty: response_header.is_empty
		end

	send_header is
			-- Send `response_header' to client.
			-- If there is an error sending `s' to the client, the
			-- connection will be closed.
		require
			header_not_send: not is_header_send
			client_connected: has_connected_client
			header_ends_with_body_separator: response_header_ends_with_body_separator
		do
			send_string (response_header)
			is_header_send := True
		ensure
			header_send: is_header_send
		end

	send_string (s: STRING) is
			-- Write `s' to `client'.
			-- Using this routine has the advantage that output to client
			-- can be written to `stderr', if you desire to do so.
			-- If there is an error sending `s' to the client, the
			-- connection will be closed.
		require
			client_connected: has_connected_client
		do
			if s /= Void and then not s.is_empty then
				debug ("httpd")
					if s.count <= 2048 then
						stderr.write_string (s)
					else
						stderr.write_string (s.substring (1, 2048))
						stderr.write_string ("...%N")
					end
				end
				client.write_string (s)
				if client.errno.is_not_ok then
					client.close
				end
			end
		end

	add_field (a_field_name, a_content: STRING) is
			-- Add a MIME field to `report_header'.
		require
			header_not_ended: not response_header_ends_with_body_separator
			a_field_name_not_empty: a_field_name /= Void and then not a_field_name.is_empty
		do
			response_header.append_string (a_field_name)
			response_header.append_string (once_colon)
			if a_content /= Void and then not a_content.is_empty then
				response_header.append_string (a_content)
			end
			add_new_line
		end

	add_new_line is
			-- Add a new line to `response_header'.
		require
			header_not_ended: not response_header_ends_with_body_separator
		do
			response_header.append_character ('%R')
			response_header.append_character ('%N')
		end


feature {NONE} -- Conversion of MIME body to form fields

	mime_body_to_form_fields is
			-- Convert a multipart/form-data body to key/value pairs.
			-- If a body is not of the proper format, it is not returned.
		require
			request_body_set: request_body /= Void
		local
			equality_tester: UC_STRING_EQUALITY_TESTER
		do
			if request_body.is_multipart then
				request_form_fields := url_encoder.mime_encoded_to_field_name_value_pair (Void, request_body)
			else
				create request_form_fields.make_default
				create equality_tester
				request_form_fields.set_key_equality_tester (equality_tester)
			end
		ensure
			request_form_fields_not_void: request_form_fields /= Void
			not_too_many_request_form_fields: request_form_fields.count <= request_body.parts_count
		end

	query_to_form_fields (a_query: STRING)	is
			-- Translate the query part of an URL to `request_form_fields'.
		local
			equality_tester: UC_STRING_EQUALITY_TESTER
		do
			if a_query /= Void then
				request_form_fields := url_encoder.url_encoded_to_field_name_value_pair (a_query)
			else
				create request_form_fields.make_default
				create equality_tester
				request_form_fields.set_key_equality_tester (equality_tester)
			end
		ensure
			request_form_fields_not_void: request_form_fields /= Void
		end

	urlencoded_to_form_fields is
			-- Convert a form-urlencoded body to key/value pairs.
		require
			request_body_set: request_body /= Void
			not_multipart: not request_body.is_multipart
		do
			request_form_fields := url_encoder.url_encoded_to_field_name_value_pair (request_body.as_string)
		ensure
			request_form_fields_not_void: request_form_fields /= Void
		end


feature {NONE} -- Implementation

	default_html_page_name: STRING is "index.html"
			-- Page retrieved when a directory is passed as resource.

	dump_request is
			-- Debug helper, dumps raw client request to stdout.
		require
			client_connected: has_connected_client
		do
			from
				client.read_character
			until
				client.eof
			loop
				fd_stdout.write_character (client.last_character)
				client.read_character
			end
		end

	remap_http_method (method: STRING): STRING is
			-- Returns the HTTP method to use if `method' is POST and one
			-- of the variables starts with "http-method:". The string
			-- after "http-method:" is used as the method the user
			-- actually wanted to sent. Useful for browsers that can send
			-- only GET/POST.
		require
			method_not_empty: method /= Void and then not method.is_empty
			request_form_fields_not_void: request_form_fields /= Void
		local
			value: EPX_KEY_VALUE
			p: INTEGER
		do
			if method.is_equal (http_method_POST) then
				from
					request_form_fields.start
				until
					Result /= Void or else
					request_form_fields.after
				loop
					value := request_form_fields.item_for_iteration
					p := value.key.substring_index (once_http_method, 1)
					if p = 1 then
						Result := value.key.substring (p + once_http_method.count, value.key.count)
					end
					request_form_fields.forth
				end
			end
			if Result = Void then
				Result := method
			end
		ensure
			remapped_method_not_empty: Result /= Void and then not Result.is_empty
		end

	select_readable: EPX_SELECT is
			-- Test if client socket becomes readable.
		once
			create Result.make
		ensure
			select_readable_not_void: Result /= Void
		end

	select_timeout: EPX_TIME_VALUE is
			-- Immediate timeout for `select_readable'
		do
			create Result.make
		ensure
			select_timeout_not_void: Result /= Void
			immediate_timeout: Result.microseconds = 0 and Result.seconds = 0
		end

	socket: EPX_TCP_SERVER_SOCKET
			-- Listening socket.

	url_encoder: expanded EPX_URL_ENCODING
			-- Decoding of encoded passed values


feature {NONE} -- Once STRINGs, Eiffel's worst feature

	once_100_continue: STRING is "100-continue"
	once_close: STRING is "close"
	once_colon: STRING is ": "
	once_file_url_prefix: STRING is "file://localhost"
	once_http_method: STRING is "http-method:"
	once_http_url_prefix: STRING is "http://localhost"
	once_http_version: STRING is "HTTP/1.1 "
	once_identity: STRING is "identity"
	once_mime_version: STRING is "1.0"
	once_none: STRING is "none"
	once_opera_7: STRING is "Opera/7."

feature {NONE} -- File extensions

	once_css: STRING is ".css"
	once_js: STRING is ".js"
	once_png: STRING is ".png"
	once_xhtml: STRING is ".xhtml"
	once_xml: STRING is ".xml"
	once_xsl: STRING is ".xsl"
	once_xslt: STRING is ".xslt"


invariant

	root_is_void_or_not_empty: root /= Void implies not root.is_empty
	root_is_absolute: root /= Void implies root.is_absolute
	root_is_resolved: root /= Void implies root.is_resolved
	root_not_ends_with_slash: root /= Void implies root.item (root.count) /= '/'
	-- Check is expensive at run-time:
	--root_is_void_or_exists: root /= Void implies is_directory (root)
	root_url_not_void: root_url /= Void
	root_url_is_absolute: root_url.is_absolute
	socket_not_void: socket /= Void implies socket.is_open
	socket_non_blocking: socket /= Void implies not socket.is_blocking_io
	client_http_version_known: has_connected_client implies client_http_version /= Void and then not client_http_version.is_empty
	body_send_implies_header_send: is_body_send implies is_header_send

	dynamic_resources_not_void: dynamic_resources /= Void
	fixed_resources_not_void: fixed_resources /= Void
	resources_not_void: resources /= Void

	response_header_not_void: response_header /= Void

end
